#!/usr/bin/env -S guile -e main -s
coding: utf-8
!#

(use-modules
 (ice-9 match)
 (ice-9 peg)
 (ice-9 rdelim)
 (juniority grammar)
 )

;; This program transforms an idol_database
;; idol/movie line into a statement in Juniority
;; format.
;; Consequently, it can be piped into
;; `juniority-import'.
;;
;; https://github.com/ikki123-aidoru/idol_database

(define-peg-pattern DASH none "-")
(define-peg-pattern SLASH none "/")
(define-peg-pattern digit body (range #\0 #\9))
(define-peg-pattern year all (+ digit))
(define-peg-pattern month all (+ digit))
(define-peg-pattern day all (+ digit))
(define-peg-pattern dash-date all
  (and year DASH month DASH day))
(define-peg-pattern slash-date all
  (and day SLASH month SLASH year))
(define-peg-pattern date all
  (or dash-date slash-date "N/A"))


(define (main args)
  (match args
    ((bin "help"   . other) (display-help bin))
    ((bin "-h"     . other) (display-help bin))
    ((bin "--help" . other) (display-help bin))
    ((bin idols-file movies-file)
     (let ((idols-hash (main-idols idols-file)))
       (main-movies idols-hash movies-file)))
    ((bin . other) (display-help bin))
    ))
(define (display-help bin)
  (format (current-error-port) "Usage: ~a path/to/idols.tsv path/to/movies.tsv

This utility is meant to be used in concert with `juniority-import`. E.g.:
    ~a path/to/idols.tsv path/to/movies.tsv | juniority-import path/to/juniority.db
" bin bin))

(define (main-idols idols-file)
  (let ((idols-hash (make-hash-table))
        (idols-port (open-file idols-file "r"))
        )
    (do ((line (read-line idols-port) (read-line idols-port)))
        ((eof-object? line))

      (match (string-split line #\tab)
        ;; Ignore header
        (("Name" "Japanese Name" "Alias" "DOB" "Doubtful DOB") (noop))

        ;; Parse data
        ((en-name "" alias dob _doubtful-dob) (noop)) ;; no
        ((en-name jp-name alias dob _doubtful-dob)
         (let ((date-of-birth
                (if (or (string-null? dob)
                        (string-any
                         (lambda (ch) (not (or (char-numeric? ch) (eqv? #\- ch))))
                         dob))
                    ""
                    (normalize-date dob))))
           (format #t "@'~a' '~a'~a\n"
                   jp-name
                   en-name
                   date-of-birth)
           (when (not (string-null? alias))
             (format #t "@'~a' '~a'~a\n"
                     alias
                     en-name
                     date-of-birth))
           (hash-set! idols-hash en-name jp-name)
           ))

        ;; Ignore empty lines
        (() (noop))
        ))
    idols-hash))

(define (main-movies idols-hash movies-file)
  (let ((movies-port (open-file movies-file "r")))
    (do ((line (read-line movies-port) (read-line movies-port)))
        ((eof-object? line))

      (match (string-split line #\tab)
        ;; Ignore headers
        (("Title" "Series" "Maker" "Release Date" "Actress") (noop))

        ;; Parse data
        ((ident series _maker published idol-en-name)
         (let ((idol-jp-name (hash-ref idols-hash idol-en-name))
               (date-of-publication
                (if (or (string-null? published)
                        (string-any
                         (lambda (ch) (not (or (char-numeric? ch) (eqv? #\- ch))))
                         published))
                    ""
                    (normalize-date published))))
           (if idol-jp-name
               (format #t "~a @'~a' ~a\n"
                       ident
                       idol-jp-name
                       date-of-publication)
               (noop))))

        ;; Ignore empty lines
        (() (noop))
        ))))

(define (normalize-date the-date)
  (match (peg:tree (match-pattern date the-date))
    (('date "N/A") "")
    (('date ('dash-date _ _ _))
     (format #f " ~a" the-date))
    (('date ('slash-date ('day day) ('month month) ('year year)))
     (format #f " ~a-~a-~a"
             (if (eqv? 2 (string-length year))
                 (if (< (string->number year) 35)
                     (string-append "20" year)
                     (string-append "19" year))
                 year)
             month
             day))))
