CREATE TABLE IF NOT EXISTS tag (
id   INTEGER PRIMARY KEY AUTOINCREMENT,
name VARCHAR(255) NOT NULL
);
CREATE UNIQUE INDEX IF NOT EXISTS tag_name_unique ON tag (name);

CREATE TABLE IF NOT EXISTS note (
id      INTEGER PRIMARY KEY AUTOINCREMENT,
text    TEXT NOT NULL,
created INTEGER NOT NULL, -- epoch
updated INTEGER NOT NULL  -- epoch
);

CREATE TABLE IF NOT EXISTS studio (
id        INTEGER PRIMARY KEY AUTOINCREMENT,
call_sign VARCHAR(255) NOT NULL,
full_name VARCHAR(255),
rating    SMALLINT
);
CREATE UNIQUE INDEX IF NOT EXISTS studio_call_sign_unique ON studio (call_sign);
CREATE INDEX IF NOT EXISTS studio_full_name ON studio (full_name);

CREATE TABLE IF NOT EXISTS idol (
id             INTEGER PRIMARY KEY AUTOINCREMENT,
jp_name        VARCHAR(255),
birth_date     VARCHAR(10),
rating         SMALLINT
);
CREATE UNIQUE INDEX IF NOT EXISTS idol_jp_name_unique ON idol (jp_name);

CREATE TABLE IF NOT EXISTS volume (
id         INTEGER PRIMARY KEY AUTOINCREMENT,
identifier VARCHAR(255) NOT NULL,
title      VARCHAR(255),
rating     SMALLINT,
published  VARCHAR(10) -- YYYY-MM-DD
);
CREATE UNIQUE INDEX IF NOT EXISTS volume_identifier_unique ON volume (identifier);
CREATE INDEX IF NOT EXISTS volume_title ON volume (title);
CREATE INDEX IF NOT EXISTS volume_published ON volume (published);

CREATE TABLE IF NOT EXISTS volume_idol (
id        INTEGER PRIMARY KEY AUTOINCREMENT,
volume_id INTEGER,
idol_id   INTEGER,
CONSTRAINT fk_volume_idol_volume
FOREIGN KEY (volume_id) REFERENCES volume (id)
ON DELETE RESTRICT
ON UPDATE RESTRICT,
CONSTRAINT fk_volume_idol_idol
FOREIGN KEY (idol_id) REFERENCES idol (id)
ON DELETE RESTRICT
ON UPDATE RESTRICT
);
CREATE INDEX IF NOT EXISTS volume_idol_volume ON volume_idol (volume_id);
CREATE INDEX IF NOT EXISTS volume_idol_idol ON volume_idol (idol_id);
CREATE UNIQUE INDEX IF NOT EXISTS volume_idol_unique ON volume_idol (volume_id, idol_id);

CREATE TABLE IF NOT EXISTS volume_studio (
id        INTEGER PRIMARY KEY AUTOINCREMENT,
volume_id INTEGER,
studio_id INTEGER,
CONSTRAINT fk_volume_studio_volume
FOREIGN KEY (volume_id) REFERENCES volume (id)
ON DELETE RESTRICT
ON UPDATE RESTRICT,
CONSTRAINT fk_volume_studio_studio
FOREIGN KEY (studio_id) REFERENCES studio (id)
ON DELETE RESTRICT
ON UPDATE RESTRICT
);
CREATE INDEX IF NOT EXISTS volume_studio_volume ON volume_studio (volume_id);
CREATE INDEX IF NOT EXISTS volume_studio_studio ON volume_studio (studio_id);
CREATE UNIQUE INDEX IF NOT EXISTS volume_studio_unique ON volume_studio (volume_id, studio_id);

CREATE TABLE IF NOT EXISTS volume_tag (
id        INTEGER PRIMARY KEY AUTOINCREMENT,
volume_id INTEGER,
tag_id    INTEGER,
CONSTRAINT fk_volume_tag_volume
FOREIGN KEY (volume_id) REFERENCES volume (id)
ON DELETE RESTRICT
ON UPDATE RESTRICT,
CONSTRAINT fk_volume_tag_tag
FOREIGN KEY (tag_id) REFERENCES tag (id)
ON DELETE RESTRICT
ON UPDATE RESTRICT
);
CREATE INDEX IF NOT EXISTS volume_tag_volume ON volume_tag (volume_id);
CREATE INDEX IF NOT EXISTS volume_tag_tag ON volume_tag (tag_id);
CREATE UNIQUE INDEX IF NOT EXISTS volume_tag_unique ON volume_tag (volume_id, tag_id);

CREATE TABLE IF NOT EXISTS idol_tag (
id        INTEGER PRIMARY KEY AUTOINCREMENT,
idol_id   INTEGER,
tag_id    INTEGER,
CONSTRAINT fk_idol_tag_idol
FOREIGN KEY (idol_id) REFERENCES idol (id)
ON DELETE RESTRICT
ON UPDATE RESTRICT,
CONSTRAINT fk_idol_tag_tag
FOREIGN KEY (tag_id) REFERENCES tag (id)
ON DELETE RESTRICT
ON UPDATE RESTRICT
);
CREATE INDEX IF NOT EXISTS idol_tag_tag ON idol_tag (tag_id);
CREATE INDEX IF NOT EXISTS idol_tag_idol ON idol_tag (idol_id);
CREATE UNIQUE INDEX IF NOT EXISTS idol_tag_unique ON idol_tag (idol_id, tag_id);

CREATE TABLE IF NOT EXISTS studio_tag (
id        INTEGER PRIMARY KEY AUTOINCREMENT,
studio_id INTEGER,
tag_id    INTEGER,
CONSTRAINT fk_studio_tag_studio
FOREIGN KEY (studio_id) REFERENCES studio (id)
ON DELETE RESTRICT
ON UPDATE RESTRICT,
CONSTRAINT fk_studio_tag_tag
FOREIGN KEY (tag_id) REFERENCES tag (id)
ON DELETE RESTRICT
ON UPDATE RESTRICT
);
CREATE INDEX IF NOT EXISTS studio_tag_tag ON studio_tag (tag_id);
CREATE INDEX IF NOT EXISTS studio_tag_studio ON studio_tag (studio_id);
CREATE UNIQUE INDEX IF NOT EXISTS studio_tag_unique ON studio_tag (studio_id, tag_id);

CREATE TABLE IF NOT EXISTS volume_note (
id        INTEGER PRIMARY KEY AUTOINCREMENT,
volume_id INTEGER,
note_id   INTEGER,
CONSTRAINT fk_volume_note_volume
FOREIGN KEY (volume_id) REFERENCES volume (id)
ON DELETE RESTRICT
ON UPDATE RESTRICT,
CONSTRAINT fk_volume_note_note
FOREIGN KEY (note_id) REFERENCES note (id)
ON DELETE RESTRICT
ON UPDATE RESTRICT
);
CREATE INDEX IF NOT EXISTS volume_note_note ON volume_note (note_id);
CREATE INDEX IF NOT EXISTS volume_note_volume ON volume_note (volume_id);
CREATE UNIQUE INDEX IF NOT EXISTS volume_note_unique ON volume_note (volume_id, note_id);

CREATE TABLE IF NOT EXISTS idol_note (
id        INTEGER PRIMARY KEY AUTOINCREMENT,
idol_id   INTEGER,
note_id   INTEGER,
CONSTRAINT fk_idol_note_idol
FOREIGN KEY (idol_id) REFERENCES idol (id)
ON DELETE RESTRICT
ON UPDATE RESTRICT,
CONSTRAINT fk_idol_note_note
FOREIGN KEY (note_id) REFERENCES note (id)
ON DELETE RESTRICT
ON UPDATE RESTRICT
);
CREATE INDEX IF NOT EXISTS idol_note_note ON idol_note (note_id);
CREATE INDEX IF NOT EXISTS idol_note_idol ON idol_note (idol_id);
CREATE UNIQUE INDEX IF NOT EXISTS idol_note_unique ON idol_note (idol_id, note_id);

CREATE TABLE IF NOT EXISTS studio_note (
id        INTEGER PRIMARY KEY AUTOINCREMENT,
studio_id INTEGER,
note_id   INTEGER,
CONSTRAINT fk_studio_note_studio
FOREIGN KEY (studio_id) REFERENCES studio (id)
ON DELETE RESTRICT
ON UPDATE RESTRICT,
CONSTRAINT fk_studio_note_note
FOREIGN KEY (note_id) REFERENCES note (id)
ON DELETE RESTRICT
ON UPDATE RESTRICT
);
CREATE INDEX IF NOT EXISTS studio_note_note ON studio_note (note_id);
CREATE INDEX IF NOT EXISTS studio_note_studio ON studio_note (studio_id);
CREATE UNIQUE INDEX IF NOT EXISTS studio_note_unique ON studio_note (studio_id, note_id);

CREATE TABLE IF NOT EXISTS tag_note (
id        INTEGER PRIMARY KEY AUTOINCREMENT,
tag_id    INTEGER,
note_id   INTEGER,
CONSTRAINT fk_tag_note_tag
FOREIGN KEY (tag_id) REFERENCES tag (id)
ON DELETE RESTRICT
ON UPDATE RESTRICT,
CONSTRAINT fk_tag_note_note
FOREIGN KEY (note_id) REFERENCES note (id)
ON DELETE RESTRICT
ON UPDATE RESTRICT
);
CREATE INDEX IF NOT EXISTS tag_note_note ON tag_note (note_id);
CREATE INDEX IF NOT EXISTS tag_note_tag ON tag_note (tag_id);
CREATE UNIQUE INDEX IF NOT EXISTS tag_note_unique ON tag_note (tag_id, note_id);
